/*  SQuaLe
 *
 *  Copyright (C) 2004 Julien Moutte <julien@moutte.net>
 *
 *  squaleclient.c : Source code for SQuaLe client library.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
#include "squaleclient.h"

#include <glib.h>

#define DEFAULT_SOCKET "/tmp/squale.sock"

static gint
read_from_socket (gint fd, void *dest, gint n_bytes)
{
  gint read_bytes, read_so_far = 0;
  
  while (read_so_far < n_bytes) {
    read_bytes = read (fd, dest + read_so_far, n_bytes - read_so_far);
    if (read_bytes > 0) {
      read_so_far += read_bytes;
      if (read_so_far >= n_bytes) {
        /* We reached the required number of bytes */
        return read_so_far;
      }
    }
    else {
      if (read_bytes == 0) {
        return read_bytes;
      }
      else {
        switch (errno) {
          case EAGAIN:
            break;
          default :
            g_warning ("an error occured: %s", strerror (errno));
            break;
        }
      }
    }
  }
}

SqualeResult *
squale_run_query (const char *connection_name,
                  const char *query,
                  const char *socket_name,
                  int s_timeout)
{
  gint fd = -1, status = -1;
  fd_set set;
  struct timeval timeout;
  gint32 query_length = 0, connection_length = 0;
  gint32 assignation_time, processing_time;
  struct sockaddr_un addr;
  char buf;
  SqualeResult *result = NULL;
  
  g_return_val_if_fail (connection_name != NULL, NULL);
  g_return_val_if_fail (query != NULL, NULL);
  
  result = g_new0 (SqualeResult, 1);

  result->assignation_time = 0;
  result->processing_time = 0;
  result->num_fields = 0;
  result->num_rows = 0;
  result->column_names = NULL;
  result->data = NULL;
  result->affected_rows = 0;
  result->error = NULL;
  
  fd = socket (PF_UNIX, SOCK_STREAM, 0);
  
  /* Clear the struct */
  memset (&(addr), '\0', sizeof (addr));
  
  /* Unix socket */
  addr.sun_family = AF_UNIX;
  
  /* Set filename */
  if (socket_name == NULL) {
    strncpy (addr.sun_path, DEFAULT_SOCKET, sizeof (addr.sun_path) - 1);
  }
  else {
    strncpy (addr.sun_path, socket_name, sizeof (addr.sun_path) - 1);
  }
  
  if (connect (fd, (struct sockaddr*) &addr, sizeof (addr)) < 0) {
    result->status = SQUALE_RESULT_ERROR;
    result->error = g_strdup_printf ("Failed connecting to socket '%s': %s",
                                     addr.sun_path, strerror (errno));
    close (fd);
    return result;
  }
  
  connection_length = strlen (connection_name);
  query_length = strlen (query);
  
  /* Initialize the file descriptor set. */
  FD_ZERO (&set);
  FD_SET (fd, &set);
  
  write (fd, &connection_length, sizeof (gint32));
  write (fd, connection_name, connection_length);
  write (fd, &query_length, sizeof (gint32));
  write (fd, query, query_length);
  
  if (s_timeout > 0) {
    timeout.tv_sec = s_timeout;
    timeout.tv_usec = 0;
    status = select (FD_SETSIZE, &set, NULL, NULL, &timeout);
  }
  else {
    status = select (FD_SETSIZE, &set, NULL, NULL, NULL);
  }
  
  if (status < 0) {
    result->status = SQUALE_RESULT_ERROR;
    result->error = g_strdup_printf ("An error occured while waiting for " \
                                     "input on socket '%s': %s",
                                     addr.sun_path, strerror (errno));
    close (fd);
    return result;
  }
  else if (status == 0) {
    result->status = SQUALE_RESULT_ERROR;
    result->error = g_strdup_printf ("Timeout reached while waiting for " \
                                     "input on socket '%s'", addr.sun_path);
    close (fd);
    return result;
  }
  
  read_from_socket (fd, &assignation_time, sizeof (gint32));
  read_from_socket (fd, &processing_time, sizeof (gint32));
  read_from_socket (fd, &buf, sizeof (char));
  
  result->assignation_time = assignation_time;
  result->processing_time = processing_time;
  
  if (buf == 'E') {
    gint32 error_length;
    gint read_bytes;
    char *error_message = NULL;
    
    result->status = SQUALE_RESULT_ERROR;
    
    read_bytes = read_from_socket (fd, &error_length, sizeof (gint32));
    if (read_bytes == sizeof (gint32)) {
      if (error_length == 0) {
        result->error = g_strdup ("Empty error message");
      }
      else {
        error_message = g_malloc0 (error_length + 1);
        read_bytes = read_from_socket (fd, error_message, error_length);
        if (read_bytes == error_length) {
          result->error = error_message;
          close (fd);
        }
        else {
          result->status = SQUALE_RESULT_ERROR;
          result->error = g_strdup_printf ("Failed reading error message " \
                                           "from socket: %s", strerror (errno));
          close (fd);
          return result;
        }
      }
    }
    else {
      result->status = SQUALE_RESULT_ERROR;
      result->error = g_strdup_printf ("Failed reading error length " \
                                       "from socket: %s", strerror (errno));
      close (fd);
      return result;
    }
  }
  else if (buf == 'A') {
    gint32 affected_rows;
    gint read_bytes;
    
    result->status = SQUALE_RESULT_AFFECTED_ROWS;
    
    read_bytes = read_from_socket (fd, &affected_rows, sizeof (gint32));
    if (read_bytes == sizeof (gint32)) {
      result->affected_rows = affected_rows;
      close (fd);
    }
    else {
      result->status = SQUALE_RESULT_ERROR;
      result->error = g_strdup_printf ("Failed reading affected rows " \
                                       "from socket: %s", strerror (errno));
      close (fd);
      return result;
    }
  }
  else if (buf == 'R' || buf == 'W') {
    gint32 num_fields, i;
    gulong num_rows, j;
    gint read_bytes;
    
    if (buf == 'R') {
      result->status = SQUALE_RESULT_RESULTSET;
    }
    else {
      result->status = SQUALE_RESULT_RESULTSET_WITH_WARNING;
    }
    
    /* Reading number of fields */
    read_bytes = read_from_socket (fd, &num_fields, sizeof (gint32));
    if (read_bytes == sizeof (gint32)) {
      result->num_fields = num_fields;
      if (num_fields != 0) {
        result->column_names = g_malloc0 (num_fields * sizeof (char *));
      }
      else {
        result->column_names = NULL;
      }
    
      for (i = 0; i < num_fields; i++) {
        gint32 field_length;
        char *field_value = NULL;
      
        read_bytes = read_from_socket (fd, &field_length, sizeof (gint32));
        if (read_bytes == sizeof (gint32)) {
          if (field_length == 0) {
            /* Handling zero length fields */
            result->column_names[i] = NULL;
          }
          else {
            field_value = g_malloc0 (field_length + 1);
            read_bytes = read_from_socket (fd, field_value, field_length);
            if (read_bytes == field_length) {
              result->column_names[i] = field_value;
            }
            else {
              result->status = SQUALE_RESULT_ERROR;
              result->error = g_strdup_printf ("Failed reading field " \
                                               "value from socket " \
                                               "(expected %d got %d " \
                                               "bytes): %s", field_length,
                                               read_bytes, strerror (errno));
              close (fd);
              return result;
            }
          }
        }
        else {
          result->status = SQUALE_RESULT_ERROR;
          result->error = g_strdup_printf ("Failed reading field " \
                                           "length from socket " \
                                           "(expected %d got %d " \
                                           "bytes): %s", sizeof (gint32),
                                           read_bytes, strerror (errno));
          close (fd);
          return result;
        }
      }
    
      read_bytes = read_from_socket (fd, &num_rows, sizeof (gulong));
      if (read_bytes == sizeof (gulong)) {
        result->num_rows = num_rows;
        if (num_rows != 0) {
          result->data = g_malloc0 (num_rows * sizeof (char **));
        }
        else {
          result->data = NULL;
        }
    
        for (j = 0; j < num_rows; j++) {
          result->data[j] = g_malloc0 (num_fields * sizeof (char *));
        }
    
        for (j = 0; j < num_rows; j++) {
          for (i = 0; i< num_fields; i++) {
            gint32 field_length;
            char *field_value = NULL;
        
            read_bytes = read_from_socket (fd, &field_length, sizeof (gint32));
            if (read_bytes == sizeof (gint32)) {
              if (field_length == 0) {
                /* Handling zero length fields */
                result->data[j][i] = NULL;
              }
              else {
                field_value = g_malloc0 (field_length + 1);
                read_bytes = read_from_socket (fd, field_value, field_length);
                if (read_bytes == field_length) {
                  result->data[j][i] = field_value;
                }
                else {
                  result->status = SQUALE_RESULT_ERROR;
                  result->error = g_strdup_printf ("Failed reading field " \
                                                   "value from socket " \
                                                   "(expected %d got %d " \
                                                   "bytes): %s", field_length,
                                                   read_bytes, strerror (errno));
                  close (fd);
                  return result;
                }
              }
            }
            else {
              result->status = SQUALE_RESULT_ERROR;
              result->error = g_strdup_printf ("Failed reading field " \
                                               "length from socket " \
                                               "(expected %d got %d " \
                                               "bytes): %s", sizeof (gint32),
                                               read_bytes, strerror (errno));
              close (fd);
              return result;
            }
          }
        }
      }
      else {
        result->status = SQUALE_RESULT_ERROR;
        result->error = g_strdup_printf ("Failed reading number of " \
                                         "rows from socket: %s",
                                         strerror (errno));
        close (fd);
        return result;
      }
    }
    else {
      result->status = SQUALE_RESULT_ERROR;
      result->error = g_strdup_printf ("Failed reading number of " \
                                       "fields from socket: %s",
                                       strerror (errno));
      close (fd);
      return result;
    }
    
    if (buf == 'W') { /* Read the warning message */
      gint32 field_length;
      char *field_value = NULL;
      read_bytes = read_from_socket (fd, &field_length, sizeof (gint32));
      if (read_bytes == sizeof (gint32)) {
        if (field_length == 0) {
          /* Handling zero length warning */
          result->warning = NULL;
        }
        else {
          field_value = g_malloc0 (field_length + 1);
          read_bytes = read_from_socket (fd, field_value, field_length);
          if (read_bytes == field_length) {
            result->warning = field_value;
          }
          else {
            result->status = SQUALE_RESULT_ERROR;
            result->error = g_strdup_printf ("Failed reading warning " \
                                             "value from socket " \
                                             "(expected %d got %d " \
                                             "bytes): %s", field_length,
                                             read_bytes, strerror (errno));
            close (fd);
            return result;
          }
        }
      }
      else {
        result->status = SQUALE_RESULT_ERROR;
        result->error = g_strdup_printf ("Failed reading warning length " \
                                         "from socket: %s", strerror (errno));
        close (fd);
        return result;
      }
    }
    
    close (fd);
  }
  
  return result;
}

int
squale_free_result (SqualeResult *result)
{
  g_return_val_if_fail (result != NULL, FALSE);
  
  if (result->column_names) {
    gint i;
    for (i = 0; i < result->num_fields; i++) {
      g_free (result->column_names[i]);
      result->column_names[i] = NULL;
    }
    g_free (result->column_names);
  }
  
  if (result->data) {
    gint i;
    glong j;
    for (j = 0; j < result->num_rows; j++) {
      for (i = 0; i < result->num_fields; i++) {
        if (result->data[j][i]) {
          g_free (result->data[j][i]);
          result->data[j][i] = NULL;
        }
      }
      g_free (result->data[j]);
      result->data[j] = NULL;
    }
    g_free (result->data);
    result->data = NULL;
  }
  
  if (result->error) {
    g_free (result->error);
    result->error = NULL;
  }
  
  if (result->warning) {
    g_free (result->warning);
    result->warning = NULL;
  }
  
  g_free (result);
  
  return TRUE;
}
