/*  SQuaLe
 *
 *  Copyright (C) 2005 Julien Moutte <julien@moutte.net>
 *
 *  squalepgsqlworker.c : Source for SqualePgsqlWorker object.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "squalepgsqlworker.h"
#include "squale-i18n.h"

#ifdef HAVE_DMALLOC
#include <dmalloc.h>
#endif

enum
{
  PROP_0,
  PROP_HOST,
  PROP_PORT,
  PROP_USER,
  PROP_PASSWD,
  PROP_DBNAME
};

static SqualeWorkerClass *parent_class = NULL;
 
/* ============================================================= */
/*                                                               */
/*                       Private Methods                         */
/*                                                               */
/* ============================================================= */

static GQuark
squale_pgsql_worker_error_quark (void)
{
  static GQuark quark = 0;
  if (quark == 0)
    quark = g_quark_from_static_string ("SQuaLe-PostGreSQL");
  return quark;
}

static gboolean
squale_pgsql_worker_store_resultset (SqualePgsqlWorker *pg_worker,
                                     SqualeJob *job, PGresult *result)
{
  gint32 num_fields = 0, i;
  gulong offset = 0, num_rows = 0, j;
  
  g_return_val_if_fail (SQUALE_IS_PGSQL_WORKER (pg_worker), FALSE);
  g_return_val_if_fail (SQUALE_IS_JOB (job), FALSE);
  g_return_val_if_fail (result != NULL, FALSE);
  
  num_fields = (gint32) PQnfields (result);
  num_rows = (gulong) PQntuples (result);
  
  /* Let's allocate the needed memory in the resultset data buffer.
     We need 2 gint32 and a char for the header then we can start packing
     our data */
  job->resultset->allocated_memory = SQUALE_PAGE_SIZE;
  job->resultset->data = g_malloc0 (job->resultset->allocated_memory);
  
  /* Packing number of fields */
  squale_check_mem_block (&(job->resultset->data),
                          &(job->resultset->allocated_memory), offset,
                          3 * sizeof (gint32) + sizeof (char));
  
  offset = 2 * sizeof (gint32) + sizeof (char);
  
  *(gint32 *)(job->resultset->data + offset) = num_fields;
  offset += sizeof (gint32);
  
  /* Packing columns names */
  for (i = 0; i < num_fields; i++) {
    gint32 field_length = strlen (PQfname (result, i));
    squale_check_mem_block (&(job->resultset->data),
                            &(job->resultset->allocated_memory), offset,
                            sizeof (gint32) + field_length);
    *(gint32 *)(job->resultset->data + offset) = field_length;
    offset += sizeof (gint32);
    memcpy (job->resultset->data + offset, PQfname (result, i), field_length);
    offset += field_length;
  }
  
  /* Packing number of rows */
  *(gulong *)(job->resultset->data + offset) = num_rows;
  offset += sizeof (gulong);
  
  /* Packing data row by row */
  for (j = 0; j < num_rows; j++) {
    for (i = 0; i < num_fields; i++) {
      gint32 field_length = strlen (PQgetvalue (result, j, i));
      squale_check_mem_block (&(job->resultset->data),
                              &(job->resultset->allocated_memory),
                              offset, sizeof (gint32) + field_length);
      *(gint32 *)(job->resultset->data + offset) = field_length;
      offset += sizeof (gint32);
      memcpy (job->resultset->data + offset,
              PQgetvalue (result, j, i), field_length);
      offset += field_length; 
    }
  }
  
  job->resultset->data_size = offset;
  
  return TRUE;
}

static void
squale_pgsql_notice_processor (gpointer user_data, const char *message)
{
  SqualeJob *job = NULL;
  
  g_return_if_fail (SQUALE_IS_JOB (user_data));
  
  job = SQUALE_JOB (user_data);
  
  if (message) {
    GError *warning = g_error_new (squale_pgsql_worker_error_quark (), 0,
                                   "%s", message);
    squale_job_set_warning (job, warning);
  }
}

static gboolean
squale_pgsql_worker_connect (SqualeWorker *worker)
{
  SqualePgsqlWorker *pg_worker = NULL;
  SqualeJobList *worker_joblist = NULL;
  char *conn_string = NULL, *joblist_name = NULL;
  guint nb_tries = 1;
  gboolean connected = FALSE;
  
  g_return_val_if_fail (SQUALE_IS_PGSQL_WORKER (worker), FALSE);
  
  pg_worker = SQUALE_PGSQL_WORKER (worker);
  worker_joblist = squale_worker_get_joblist (worker);
  if (SQUALE_IS_JOBLIST (worker_joblist)) {
    joblist_name = squale_joblist_get_name (worker_joblist);
    g_object_unref (worker_joblist);
    worker_joblist = NULL;
  }
  
  squale_worker_set_status (worker, _("Connecting"));
  
  conn_string = g_strdup_printf ("host=%s port=%s dbname=%s user=%s password=%s",
                                 pg_worker->host, pg_worker->port,
                                 pg_worker->dbname, pg_worker->user,
                                 pg_worker->passwd);
  
  while (!connected && (nb_tries <= 3)) {
    
    g_message (_("Joblist '%s': PostGreSQL worker (%p) trying to establish " \
               "connection to %s (attempt %d)"), joblist_name, worker,
               pg_worker->dbname, nb_tries);
    
    pg_worker->conn = PQconnectdb (conn_string);
  
    if (PQstatus (pg_worker->conn) != CONNECTION_OK) {
      g_warning (_("Joblist '%s': Connection attempt %d to PostGreSQL " \
                 "database %s failed for worker (%p): %s"), joblist_name,
                 nb_tries, pg_worker->dbname, worker,
                 PQerrorMessage (pg_worker->conn));
      PQfinish (pg_worker->conn);
      pg_worker->conn = NULL;
    }
    else {
      g_message (_("Joblist '%s': PostGreSQL worker (%p) successfully " \
                 "connected to %s"), joblist_name, worker, pg_worker->dbname);
      connected = TRUE;
    }
    nb_tries++;
  }
  
  if (conn_string)
    g_free (conn_string);
  if (joblist_name)
    g_free (joblist_name);
  
  return connected;
}

static gboolean
squale_pgsql_worker_disconnect (SqualeWorker *worker)
{
  SqualePgsqlWorker *pg_worker = NULL;
  SqualeJobList *worker_joblist = NULL;
  char *joblist_name = NULL;
  
  g_return_val_if_fail (SQUALE_IS_PGSQL_WORKER (worker), FALSE);
  
  pg_worker = SQUALE_PGSQL_WORKER (worker);
  worker_joblist = squale_worker_get_joblist (worker);
  if (SQUALE_IS_JOBLIST (worker_joblist)) {
    joblist_name = squale_joblist_get_name (worker_joblist);
    g_object_unref (worker_joblist);
    worker_joblist = NULL;
  }
  
  squale_worker_set_status (worker, _("Disconnecting"));
  
  g_message (_("Joblist '%s': PostGreSQL worker (%p) shutting down " \
             "connection to %s"), joblist_name, worker, pg_worker->dbname);
  
  PQfinish (pg_worker->conn);
  pg_worker->conn = NULL;
  
  if (joblist_name)
    g_free (joblist_name);
  
  return TRUE;
}

static gpointer
squale_pgsql_worker_run (gpointer worker)
{
  SqualePgsqlWorker *pg_worker = NULL;
  SqualeJobList *joblist = NULL;
  SqualeJob *job = NULL;
  
  g_return_val_if_fail (SQUALE_IS_PGSQL_WORKER (worker), NULL);
  
  pg_worker = SQUALE_PGSQL_WORKER (worker);
  joblist = SQUALE_WORKER (pg_worker)->joblist;
  
  squale_worker_set_running (SQUALE_WORKER (pg_worker), TRUE);
  
  if (!squale_worker_connect (SQUALE_WORKER (pg_worker))) {
    squale_worker_set_running (SQUALE_WORKER (pg_worker), FALSE);
    return NULL;
  }
  
  squale_worker_set_status (SQUALE_WORKER (pg_worker), _("Sleeping"));
  
  /* The worker looping forever until shutdown is requested */
  while (!squale_worker_check_shutdown (SQUALE_WORKER (pg_worker))) {
    
    if (job == NULL) {
      job = squale_joblist_assign_pending_job (joblist, FALSE);
    }
    
    if (SQUALE_IS_JOB (job)) {
      PGresult *res;
      
      g_object_ref (job);
      
      /* We set a notice processor to catch warnings and set them on the job */
      PQsetNoticeProcessor (pg_worker->conn, squale_pgsql_notice_processor,
                            job);
      
      squale_worker_set_status (SQUALE_WORKER (pg_worker), job->query);
      
      if (PQstatus (pg_worker->conn) != CONNECTION_OK) {
        g_warning (_("PostGreSQL connection to %s went down, trying to cycle"),
                   pg_worker->dbname);
        if (!squale_worker_connect (SQUALE_WORKER (pg_worker))) {
          /* Error: No connection */
          GError *error = g_error_new (squale_pgsql_worker_error_quark (), 0,
                                       _("PostGreSQL worker was unable to " \
                                         "connect to %s"), pg_worker->dbname);
          squale_job_set_error (job, error);
          SQUALE_WORKER (pg_worker)->nb_errors++;
        }
        SQUALE_WORKER (pg_worker)->nb_db_conn_cycles++;
      }
      else {
        /* We have a job assigned to us */
        res = PQexec (pg_worker->conn, job->query);
        switch (PQresultStatus (res)) {
          case PGRES_COMMAND_OK:
            /* No resultset */
            job->affected_rows = (gint32) atoi (PQcmdTuples (res));
            break;
          case PGRES_TUPLES_OK:
            /* We have a resultset */
            squale_pgsql_worker_store_resultset (pg_worker, job, res);
            break;
          case PGRES_EMPTY_QUERY:
          case PGRES_BAD_RESPONSE:
          case PGRES_NONFATAL_ERROR:
          case PGRES_FATAL_ERROR:
          default:
          {
            /* Error: Query failed */
            GError *error = g_error_new (squale_pgsql_worker_error_quark (), 0,
                                         "%s", PQerrorMessage (pg_worker->conn));
            squale_job_set_error (job, error);
            SQUALE_WORKER (pg_worker)->nb_errors++;
          }
            break;
        }
      
        PQclear (res);
      }
      
      squale_job_set_status_if_match (job, SQUALE_JOB_COMPLETE,
                                      SQUALE_JOB_PROCESSING);
      
      SQUALE_WORKER (pg_worker)->nb_jobs_processed++;
      
      g_object_unref (job);
      job = NULL;
      
      squale_worker_set_status (SQUALE_WORKER (pg_worker), _("Sleeping"));
    }
    else {
      /* No pending job in the joblist let's wait for a new job, while switching
         to waiting state we check the joblist for a pending job. If there is
         one we return it and we don't wait */
      job = squale_worker_wait (SQUALE_WORKER (pg_worker));
    }
  }
  
  squale_worker_disconnect (SQUALE_WORKER (pg_worker));
  
  squale_worker_shutdown_complete (SQUALE_WORKER (pg_worker));
  squale_worker_set_running (SQUALE_WORKER (pg_worker), FALSE);
  
  return NULL;
}

static void
squale_pgsql_worker_set_property (GObject *object, guint prop_id,
                                  const GValue *value, GParamSpec *pspec)
{
  SqualePgsqlWorker *worker = NULL;
  
  g_return_if_fail (SQUALE_IS_PGSQL_WORKER (object));
  
  worker = SQUALE_PGSQL_WORKER (object);
  
  switch (prop_id)
    {
      case PROP_HOST:
        if (worker->host)
          g_free (worker->host);
        worker->host = g_strdup (g_value_get_string (value));
        break;
      case PROP_PORT:
        if (worker->port)
          g_free (worker->port);
        worker->port = g_strdup (g_value_get_string (value));
        break;
      case PROP_DBNAME:
        if (worker->dbname)
          g_free (worker->dbname);
        worker->dbname = g_strdup (g_value_get_string (value));
        break;
      case PROP_USER:
        if (worker->user)
          g_free (worker->user);
        worker->user = g_strdup (g_value_get_string (value));
        break;
      case PROP_PASSWD:
        if (worker->passwd)
          g_free (worker->passwd);
        worker->passwd = g_strdup (g_value_get_string (value));
        break;
      default :
        G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
        break;
    }
}

static void
squale_pgsql_worker_get_property (GObject *object, guint prop_id,
                                  GValue *value, GParamSpec *pspec)
{
  SqualePgsqlWorker *worker = NULL;
  
  g_return_if_fail (SQUALE_IS_PGSQL_WORKER (object));
  
  worker = SQUALE_PGSQL_WORKER (object);
  
  switch (prop_id)
    {
      case PROP_HOST:
        g_value_set_string (value, g_strdup (worker->host));
        break;
      case PROP_PORT:
        g_value_set_string (value, g_strdup (worker->port));
        break;
      case PROP_DBNAME:
        g_value_set_string (value, g_strdup (worker->dbname));
        break;
      case PROP_USER:
        g_value_set_string (value, g_strdup (worker->user));
        break;
      case PROP_PASSWD:
        g_value_set_string (value, g_strdup (worker->passwd));
        break;
      default :
        G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
        break;
    }
}

/* =========================================== */
/*                                             */
/*              Init & Class init              */
/*                                             */
/* =========================================== */

static void
squale_pgsql_worker_dispose (GObject *object)
{
  SqualePgsqlWorker *worker = NULL;
  
  worker = SQUALE_PGSQL_WORKER (object);
  
  if (worker->host)
    g_free (worker->host);
  
  worker->host = NULL;
  
  if (worker->port)
    g_free (worker->port);
  
  worker->port = NULL;
  
  if (worker->user)
    g_free (worker->user);
  
  worker->user = NULL;
  
  if (worker->passwd)
    g_free (worker->passwd);
  
  worker->passwd = NULL;
  
  if (worker->dbname)
    g_free (worker->dbname);
  
  worker->dbname = NULL;
  
  if (G_OBJECT_CLASS (parent_class)->dispose)
    G_OBJECT_CLASS (parent_class)->dispose (object);
}

static void
squale_pgsql_worker_init (SqualePgsqlWorker *worker)
{
  worker->host = NULL;
  worker->port = NULL;
  worker->user = NULL;
  worker->passwd = NULL;
  worker->dbname = NULL;
}

static void
squale_pgsql_worker_class_init (SqualePgsqlWorkerClass *klass)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
  SqualeWorkerClass *worker_class = SQUALE_WORKER_CLASS (klass);
  
  parent_class = g_type_class_peek_parent (klass);

  gobject_class->set_property = squale_pgsql_worker_set_property;
  gobject_class->get_property = squale_pgsql_worker_get_property;
  gobject_class->dispose = squale_pgsql_worker_dispose;
  
  worker_class->connect = squale_pgsql_worker_connect;
  worker_class->disconnect = squale_pgsql_worker_disconnect;
  worker_class->run = squale_pgsql_worker_run;
  
  g_object_class_install_property (gobject_class,
                                   PROP_HOST,
                                   g_param_spec_string ("host",
                                                        "Database hostname",
                                                        "The hostname to access that database",
                                                        NULL,
                                                        G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class,
                                   PROP_PORT,
                                   g_param_spec_string ("port",
                                                        "Database port",
                                                        "The port to access that database",
                                                        NULL,
                                                        G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class,
                                   PROP_DBNAME,
                                   g_param_spec_string ("dbname",
                                                        "Database name",
                                                        "The name of that database",
                                                        NULL,
                                                        G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class,
                                   PROP_USER,
                                   g_param_spec_string ("user",
                                                        "Username",
                                                        "The username to access that database",
                                                        NULL,
                                                        G_PARAM_READWRITE));
  g_object_class_install_property (gobject_class,
                                   PROP_PASSWD,
                                   g_param_spec_string ("passwd",
                                                        "Password",
                                                        "The password to access that database",
                                                        NULL,
                                                        G_PARAM_READWRITE));
}

/* ============================================================= */
/*                                                               */
/*                       Public Methods                          */
/*                                                               */
/* ============================================================= */

/* =========================================== */
/*                                             */
/*          Object typing & Creation           */
/*                                             */
/* =========================================== */

GType
squale_pgsql_worker_get_type (void)
{
  static GType worker_type = 0;

  if (!worker_type)
    {
      static const GTypeInfo worker_info = {
        sizeof (SqualePgsqlWorkerClass),
        NULL,                   /* base_init */
        NULL,                   /* base_finalize */
        (GClassInitFunc) squale_pgsql_worker_class_init,
        NULL,                   /* class_finalize */
        NULL,                   /* class_data */
        sizeof (SqualePgsqlWorker),
        0,                      /* n_preallocs */
        (GInstanceInitFunc) squale_pgsql_worker_init,
        NULL                    /* value_table */
      };

      worker_type =
        g_type_register_static (SQUALE_TYPE_WORKER, "SqualePgsqlWorker",
                                &worker_info, (GTypeFlags) 0);
    }
  return worker_type;
}

SqualePgsqlWorker *
squale_pgsql_worker_new (void)
{
  SqualePgsqlWorker *worker = g_object_new (SQUALE_TYPE_PGSQL_WORKER, NULL);

  return worker;
}
